from asyncio import get_event_loop

from tests.client import Client
from tests.server import EngineServerManager, ServerConfig, ServerStatus, log

server_config = ServerConfig("./config.json")
esm = EngineServerManager()


async def status_change_listener(status) -> None:
    if status == ServerStatus.Running:
        client = Client(
            server_config.host,
            server_config.port,
            server_config.public_key,
            True if server_config.ssl else False,
        )
        data = await client.sendMessage({"messageOne": "testOne"})
        assert data == f"ws{client.protocol()}://"
        await client.connectToWebsocket({"messageTwo": "testTwo"})
        await client.sendWebSocketMessage({"name": "Georgia"})
        await client.disconnectWebsocket()
        esm.stop()


async def server_loop_handler() -> None:
    esm.add_listener(status_change_listener)
    esm.start()
    esm.join()
    log.info("Joining main thread...")


def test_middleware() -> None:
    loop = get_event_loop()
    loop.run_until_complete(server_loop_handler())
    log.info("Joined!")



from json import dumps, loads
from typing import Union

from nacl.encoding import Base64Encoder, HexEncoder
from nacl.public import Box, PrivateKey, PublicKey


class Nacl:
    private_key: PrivateKey

    def __init__(self, private_key: PrivateKey, encoder=HexEncoder) -> None:
        self.private_key = private_key
        self.encoder = encoder

    def _decode(self, parameter: Union[PrivateKey, PublicKey]) -> str:
        return parameter.encode(encoder=self.encoder).decode()

    def decodedPrivateKey(self) -> str:
        return self._decode(self.private_key)

    def decodedPublicKey(self) -> str:
        return self._decode(self.private_key.public_key)


def custom_loads(obj) -> any:
    if isinstance(obj, str):
        obj = f'"{obj}"'
    return loads(obj)


class MailBox:
    _private_key: PrivateKey
    _box: Box

    def __init__(self, private_key: PrivateKey, hex_public_key: str) -> None:
        self._private_key = private_key
        self._box = Box(self._private_key, PublicKey(hex_public_key, HexEncoder))

    def unbox(self, encrypted_message) -> any:
        decrypted_message = self._box.decrypt(encrypted_message, encoder=Base64Encoder)
        return custom_loads(decrypted_message)

    def box(self, message: any) -> str:
        return self._box.encrypt(dumps(message).encode(), encoder=Base64Encoder).decode()



from re import fullmatch
from typing import Tuple

from aiohttp.web import Request


def is_exclude(request: Request, exclude: Tuple) -> bool:
    for pattern in exclude:
        if fullmatch(pattern, request.path):
            return True
    return False



from inspect import signature
from logging import getLogger
from operator import itemgetter
from sys import exc_info
from traceback import format_exception
from typing import Tuple

from aiohttp import WSCloseCode
from aiohttp.typedefs import Handler, Middleware
from aiohttp.web import (
    HTTPUnauthorized,
    Request,
    Response,
    StreamResponse,
    WebSocketResponse,
    middleware,
)
from nacl.public import PrivateKey

from nacl_middleware.nacl_utils import MailBox
from nacl_middleware.utils import is_exclude

mailBoxes = {}


def nacl_middleware(
    private_key: PrivateKey,
    exclude_routes: Tuple = tuple(),
    exclude_methods: Tuple = tuple(),
    log=getLogger(),
) -> Middleware:

    def nacl_decryptor(public_key, encrypted_message) -> Tuple[any, MailBox]:
        if public_key in mailBoxes:
            my_mail_box = mailBoxes[public_key]
        else:
            my_mail_box = MailBox(private_key, public_key)
            mailBoxes[public_key] = my_mail_box

        log.debug("Decrypting message...")
        message = my_mail_box.unbox(encrypted_message)
        log.debug(f"Message {message} decrypted!")
        return message, my_mail_box

    @middleware
    async def returned_middleware(request: Request, handler: Handler) -> StreamResponse:
        if not (
            is_exclude(request, exclude_routes) or request.method in exclude_methods
        ):

            try:
                log.debug("Retrieving publicKey and encryptedMessage from message...")
                publicKey, encryptedMessage = itemgetter(
                    "publicKey", "encryptedMessage"
                )(request.query)
                log.debug(
                    f"PublicKey {publicKey} and EncryptedMessage {encryptedMessage} retrieved!"
                )

                decrypted_message, my_mail_box = nacl_decryptor(
                    publicKey, encryptedMessage
                )

                request["mail_box"] = my_mail_box
                request["decrypted_message"] = decrypted_message
            except Exception:
                the_exc_info = exc_info()
                exception_str = "".join(format_exception(*the_exc_info))
                log.debug(f"Exception body: {exception_str}")
                exception = HTTPUnauthorized(
                    reason="Failed to retrieve a valid message!", body=exception_str
                )

                # Inspect the handler's signature
                return_annotation = signature(handler).return_annotation
                if return_annotation == WebSocketResponse:
                    log.debug("WebSocketResponse hook.")
                    log.debug(f"Exception is {exception_str}")
                    socket = WebSocketResponse()
                    await socket.prepare(request)
                    await socket.close(
                        code=WSCloseCode.PROTOCOL_ERROR,
                        message="".join(format_exception(exception)),
                    )
                    return socket
                elif return_annotation == Response:
                    log.debug("Response hook.")
                    log.debug(f"headers: {exception.headers}")
                    log.debug(f"status: {exception.status}")
                    log.debug(f"reason: {exception.reason}")
                    log.debug(f"body: {exception.body}")
                    return Response(
                        headers=exception.headers,
                        status=exception.status,
                        reason=exception.reason,
                        body=exception.body,
                    )

        return await handler(request)

    return returned_middleware
